import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error, r2_score
from statsmodels.graphics.tsaplots import plot_pacf


def plot(y, model_result, suptitle=''):

    rmse = mean_squared_error(y, model_result.Z) ** 0.5
    r2 = r2_score(y, model_result.Z)

    fig, ax = plt.subplots(2, 2, figsize=(12, 8))
    fig.suptitle(f'{suptitle}', fontsize=12, fontweight='bold')

    ax[0, 0].plot(y.index, model_result.Z, label='Ti Modelled', alpha=0.75)
    ax[0, 0].plot(y, label='Ti Measured', alpha=0.75)
    if 'Te' in model_result.var.keys():
        ax[0, 0].plot(y.index, model_result.var['Te'], label='Te Modelled', alpha=0.75)
    ax[0, 0].legend()
    ax[0, 0].set_ylabel('[˚C]')
    if abs(rmse) < 100:
        ax[0, 0].set_title(f'RMSE: {rmse:.4f}')
    else:
        ax[0, 0].set_title(f'RMSE: {rmse:.4e}')
    plt.setp(ax[0, 0].get_xticklabels(), rotation=30, horizontalalignment='right')

    ax[1, 0].scatter(model_result.Z, y, alpha=0.2)
    ax[1, 0].set_xlabel('Ti Modelled [˚C]')
    ax[1, 0].set_ylabel('Ti Measured [˚C]')
    if abs(r2) < 100:
        ax[1, 0].set_title(f'$R^2$: {r2:.4f}')
    else:
        ax[1, 0].set_title(f'$R^2$: {r2:.4e}')
    plt.setp(ax[0, 1].get_xticklabels(), rotation=30, horizontalalignment='right')

    ax[0, 1].plot(y.index, y - model_result.Z, label='Ti Residual', color='black', alpha=0.75)
    ax[0, 1].legend()
    ax[0, 1].set_ylabel('[˚C]')
    ax[0, 1].set_title('Residuals')

    plot_pacf(y - model_result.Z, ax=ax[1, 1], lags=50)

    fig.tight_layout()


def plot_input_data(df):

    # keep this out of pandas plotting capabilities for compatibility
    # with the rest of the plots
    fig, ax = plt.subplots(2, 1, figsize=(12, 8))

    ax[0].plot(df.index, df['Ti'], label='Ti', alpha=0.75)
    ax[0].plot(df.index, df['Ta'], label='Ta', alpha=0.75)
    ax[0].legend()
    ax[0].set_ylabel('[˚C]')
    ax[0].set_xticks([], minor=[])
    ax[0].set_xlim(df.index[0], df.index[-1])

    ax[1].plot(df.index, df['Ph'], label='Ph', alpha=0.75, color='black')
    ax[1].legend()
    ax[1].set_ylabel('[kWh]')
    ax[1].set_xlim(df.index[0], df.index[-1])

    fig.tight_layout()




from typing import Any, Callable, Dict, List, Optional

import pandas as pd

from darkgreybox.predict import predict_models
from darkgreybox.prefit import prefit_models
from darkgreybox.train import train_models


def darkgreyfit(
    models: List[Any],
    X_train: pd.DataFrame,
    y_train: pd.Series,
    X_test: pd.DataFrame,
    y_test: pd.Series,
    ic_params_map: Dict,
    error_metric: Callable,
    prefit_splits: Optional[List] = None,
    prefit_filter: Optional[Callable] = None,
    reduce_train_results: bool = False,
    method: str = 'nelder',
    obj_func: Optional[Callable] = None,
    n_jobs: int = -1,
    verbose: int = 10
) -> pd.DataFrame:
    """
    Given a list of `models` applies a prefit according to `prefit_splits`, then fits them
    to the training data and evaluates them on the test data.

    Params:
        models: list of `model.DarkGreyModel` objects
            list of models to be trained
        X_train: `pd.DataFrame`
            A pandas DataFrame of the training input data X
        y_train: `pd.Series`
            A pandas Series of the training input data y
        X_test: `pd.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pd.Series`
            A pandas Series of the test input data y
        ic_params_map: Dict
            A dictionary of mapping functions that return the
            initial condition parameters for the test set
        error_metric: Callable
            An error metric function that confirms to the `sklearn.metrics` interface
        prefit_splits: Optional[List]
            A list of training data indices specifying sub-sections of `X_train` and `y_train`
            for the prefitting of models
        prefit_filter: Optional[Callable]
            A function acting as a filter based on the 'error' values of the trained models
        reduce_train_results: bool
            If set to True, the training dataframe will be reduced / cleaned
            by removing nan and duplicate records
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        obj_func: Optional[Callable]
            The objective function to minimise during the fitting
        n_jobs: int
            The number of parallel jobs to be run as described by `joblib.Parallel`
        verbose: int
            The degree of verbosity as described by `joblib.Parallel`

    Returns:
        `pd.DataFrame` with a record for each model's potentially viable results

    """

    models_to_train = prefit_models(
        models,
        X_train,
        y_train,
        error_metric,
        prefit_splits,
        prefit_filter,
        method,
        obj_func,
        n_jobs,
        verbose
    )

    train_df = train_models(
        models=models_to_train,
        X_train=X_train,
        y_train=y_train,
        splits=None,
        error_metric=error_metric,
        method=method,
        obj_func=obj_func,
        reduce_train_results=reduce_train_results,
        n_jobs=n_jobs,
        verbose=verbose
    )

    test_df = predict_models(
        models=train_df['model'].tolist(),
        X_test=X_test,
        y_test=y_test,
        ic_params_map=ic_params_map,
        error_metric=error_metric,
        train_results=train_df['model_result'].tolist(),
        n_jobs=n_jobs,
        verbose=verbose
    )

    return pd.concat([train_df, test_df], keys=['train', 'test'], axis=1)



from abc import ABC
from copy import deepcopy
from dataclasses import dataclass
from typing import Callable, Dict, Optional, Union

import lmfit
import numpy as np

from darkgreybox import logger


@dataclass
class DarkGreyModelResult:
    '''
    Dataclass that holds the results of the fitting of the model to a data set.
        Z: The measured variable's fit / predicted values
        X: The input values used to fit the model
        var: The variables of the model including internal ones if any
        params: The parameters of the model
    '''
    Z: np.ndarray
    X: Dict
    params: lmfit.Parameters
    var: Dict


class DarkGreyModel(ABC):
    '''
    Abstract Base Class for DarkGrey Models
    '''

    def __init__(self, params: Union[lmfit.Parameters, Dict], rec_duration: float):
        '''
        Initialises the model instance

        Parameters
        ----------
        params : dict
            A dictionary of parameters for the fitting. Key - value pairs should follow the
            `lmfit.Parameters` declaration:
            e.g. {'A' : {'value': 10, 'min': 0, 'max': 30}} - sets the initial value and the bounds
            for parameter `A`
        rec_duration : float
            The duration of each measurement record in hours
        '''

        self.result = lmfit.minimizer.MinimizerResult()

        # convert the params dict into lmfit parameters
        if isinstance(params, lmfit.Parameters):
            self.params = deepcopy(params)
        else:
            self.params = lmfit.Parameters()
            for k, v in params.items():
                self.params.add(k, **v)

        # set the number of records based on the measured variable's values
        self.rec_duration = rec_duration

    def fit(
        self,
        X: Dict,
        y: np.ndarray,
        method: str,
        ic_params: Optional[Dict] = None,
        obj_func: Optional[Callable] = None
    ):
        '''
        Fits the model by minimising the objective function value

        Parameters
        ----------
        X : dict
            A dictionary of input values for the fitting - these values are fixed during the fit.
        y : np.array
            The measured variable's values for the minimiser to fit to
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        ic_params : Optional[Dict]
            The initial condition parameters - if passed in these will overwrite
            the initial conditions in self.params
        obj_func : Optional[Callable]
            The objective function that is passed to `lmfit.minimize`/
            It must have (params, *args, **kwargs) as its method signature.
            Default: `def_obj_func`

        Returns
        -------
        `lmfit.minimizer.MinimizerResult`
            Object containing the optimized parameters and several
            goodness-of-fit statistics.
        '''

        # overwrite initial conditions
        if ic_params is not None:
            for k, v in ic_params.items():
                if k in self.params:
                    self.params[k].value = v
                else:
                    logger.warning(f'Key `{k}` not found in initial conditions params')

        # we are passing X, y to minimise as kwargs
        self.result = lmfit.minimize(
            obj_func or self.def_obj_func,
            self.params,
            kws={'model': self.model, 'X': X, 'y': y},
            method=method
        )

        self.params = self.result.params

        return self

    def predict(self, X: Dict, ic_params: Optional[Dict] = None) -> DarkGreyModelResult:
        '''
        Generates a prediction based on the result parameters and X.

        Parameters
        ----------
        X : dict
            A dictionary of input values
        ic_params : dict
            The initial condition parameters - if passed in these will overwrite
            the initial conditions in self.params

        Returns
        -------
        The results of the model
        '''

        if ic_params is not None:
            for k, v in ic_params.items():
                self.params[k].value = v

        return self.model(self.params, X)

    def model(self, params: lmfit.Parameters, X: Dict) -> DarkGreyModelResult:
        '''
        A system of differential equations describing the thermal model
        '''
        ...  # pragma: no cover

    def lock(self):
        '''
        Locks the parameters by setting `vary` to False
        '''

        for param in self.params.keys():
            self.params[param].vary = False

        return self

    @staticmethod
    def def_obj_func(params: lmfit.Parameters, *args, **kwargs):
        '''
        Default objective function
        Computes the residual between measured data and fitted data
        The model, X and y are passed in as kwargs by `lmfit.minimize`
        '''
        return ((kwargs['model'](params=params, X=kwargs['X']).Z - kwargs['y'])).ravel()



import numpy as np

from darkgreybox.base_model import DarkGreyModel, DarkGreyModelResult


class TiTeThRia(DarkGreyModel):
    '''
    A DarkGrey Model representing a TiTeThRia RC-equivalent circuit

    Notes
    -----
    See "Bacher & Madsen (2011) Identifying suitable models for the heat dynamics of buildings.
    Energy and Buildings. 43. 1511-1522. 10.1016/j.enbuild.2011.02.005." for a complete description
    of RC thermal models and the eqiuvalent circuit diagram of TiTeThRia.

    ~~~~
    # load data from e.g. pandas
    df = pd.read_csv()

    # assign internal temperature as the measured variable to be fitted to
    y = df['Internal Temperature [˚C]'].values

    # input values that will not change during the fit
    X = {
        'Ph': df['Boiler Power Output [kW]'].values,
        'Ta': df['Outside Air Temperature [˚C]'].values,
        'Th': df['Heating Circuit Temperature [˚C]'].values
    }

    # parameters to be fitted
    # 'value' - initial value
    # 'min' & 'max' - boundaries
    # 'vary' - if false, the parameter will be fixed to its initial value
    params = {
        'Ti0': {'value': y[0], 'vary': False, 'min': 15, 'max': 25},
        'Te0': {'value': y[0] - 2, 'vary': True, 'min': 10, 'max': 25},
        'Th0': {'value': y[0], 'vary': False, 'min': 10, 'max': 80},
        'Ci': {'value': 132},
        'Ce': {'value': 600},
        'Ch': {'value': 2.55, 'vary': False},
        'Rie': {'value': 0.1},
        'Rea': {'value': 1},
        'Ria': {'value': 2},
        'Rih': {'value': 0.65, 'vary': False}
    }

    # fit using the Nelder-Mead method
    model = TiTeThRia(params, rec_duration=1).fit(X, y, method='nelder')
    ~~~~
    '''

    def model(self, params, X):
        '''
        The system of differential equations describing the model

        Parameters
        ----------
        params : `lmfit.Parameters`
            - 'Ti0' : Internal temperature at t(0)
            - 'Te0' : Thermal envelope temperature at t(0)
            - 'Th0' : Heating system temperature at t(0)
            - 'Rih' : Thermal resistance between internal and heating system
            - 'Rie' : Thermal resistance between internal and thermal envelope
            - 'Rea' : Thermal resistance between thermal envelope and ambient
            - 'Ria' : Thermal resistance between internal and ambient
            - 'Ci' : Thermal capacitance of internal
            - 'Ch' : Thermal capacitance of heating system
            - 'Ce' : Thermal capacitance of thermal envelope
        X : dict
            - 'Ta' : List of ambient temperature values
            - 'Ph' : List of heating system power output values

        Returns
        -------
        Ti : np.array
            Fitted internal temperature values
        Te : np.array
            Fitted thermal envelope temperature values
        Th : np.array
            Fitted heating system temperature values
        '''

        num_rec = len(X['Ta'])

        Ti = np.zeros(num_rec)
        Te = np.zeros(num_rec)
        Th = np.zeros(num_rec)

        # alias these params/X so that the differential equations look pretty
        Ti[0] = params['Ti0']
        Te[0] = params['Te0']
        Th[0] = params['Th0']

        Rie = params['Rie'].value
        Rea = params['Rea'].value
        Rih = params['Rih'].value
        Ria = params['Ria'].value

        Ci = params['Ci'].value
        Ce = params['Ce'].value
        Ch = params['Ch'].value

        Ta = X['Ta']
        Ph = X['Ph']

        for i in range(1, num_rec):

            # the model equations
            dTi = ((Te[i-1] - Ti[i-1]) / (Rie * Ci) + (Th[i-1] - Ti[i-1]) / (Rih * Ci) +
                   (Ta[i-1] - Ti[i-1]) / (Ria * Ci)) * self.rec_duration
            dTe = ((Ti[i-1] - Te[i-1]) / (Rie * Ce) + (Ta[i-1] - Te[i-1]) / (Rea * Ce)) * self.rec_duration
            dTh = ((Ti[i-1] - Th[i-1]) / (Rih * Ch) + (Ph[i-1]) / (Ch)) * self.rec_duration

            Ti[i] = Ti[i-1] + dTi
            Te[i] = Te[i-1] + dTe
            Th[i] = Th[i-1] + dTh

        return DarkGreyModelResult(Ti, X, params, {'Ti': Ti, 'Te': Te, 'Th': Th})


class TiTeTh(DarkGreyModel):
    '''
    A DarkGrey Model representing a TiTeTh RC-equivalent circuit

    Notes
    -----
    See "Bacher & Madsen (2011) Identifying suitable models for the heat dynamics of buildings.
    Energy and Buildings. 43. 1511-1522. 10.1016/j.enbuild.2011.02.005." for a complete description
    of RC thermal models and the eqiuvalent circuit diagram of TiTeTh.

    ~~~~
    # load data from e.g. pandas
    df = pd.read_csv()

    # assign internal temperature as the measured variable to be fitted to
    y = df['Internal Temperature [˚C]'].values

    # input values that will not change during the fit
    X = {
        'Ph': df['Boiler Power Output [kW]'].values,
        'Ta': df['Outside Air Temperature [˚C]'].values,
        'Th': df['Heating Circuit Temperature [˚C]'].values
    }

    # parameters to be fitted
    # 'value' - initial value
    # 'min' & 'max' - boundaries
    # 'vary' - if false, the parameter will be fixed to its initial value
    params = {
        'Ti0': {'value': y[0], 'vary': False, 'min': 15, 'max': 25},
        'Te0': {'value': y[0] - 2, 'vary': True, 'min': 10, 'max': 25},
        'Th0': {'value': y[0], 'vary': False, 'min': 10, 'max': 80},
        'Ci': {'value': 132},
        'Ce': {'value': 600},
        'Ch': {'value': 2.55, 'vary': False},
        'Rie': {'value': 0.1},
        'Rea': {'value': 1},
        'Rih': {'value': 0.65, 'vary': False}
    }

    # fit using the Nelder-Mead method
    model = TiTeTh(params, rec_duration=1).fit(X, y, method='nelder')
    ~~~~
    '''

    def model(self, params, X):
        '''
        The system of differential equations describing the model

        Parameters
        ----------
        params : `lmfit.Parameters`
            - 'Ti0' : Internal temperature at t(0)
            - 'Te0' : Thermal envelope temperature at t(0)
            - 'Th0' : Heating system temperature at t(0)
            - 'Rih' : Thermal resistance between internal and heating system
            - 'Rie' : Thermal resistance between internal and thermal envelope
            - 'Rea' : Thermal resistance between thermal envelope and ambient
            - 'Ci' : Thermal capacitance of internal
            - 'Ch' : Thermal capacitance of heating system
            - 'Ce' : Thermal capacitance of thermal envelope
        X : dict
            - 'Ta' : List of ambient temperature values
            - 'Ph' : List of heating system power output values

        Returns
        -------
        Ti : np.array
            Fitted internal temperature values
        Te : np.array
            Fitted thermal envelope temperature values
        Th : np.array
            Fitted heating system temperature values
        '''

        num_rec = len(X['Ta'])

        Ti = np.zeros(num_rec)
        Te = np.zeros(num_rec)
        Th = np.zeros(num_rec)

        # alias these params/X so that the differential equations look pretty
        Ti[0] = params['Ti0']
        Te[0] = params['Te0']
        Th[0] = params['Th0']

        Rie = params['Rie'].value
        Rea = params['Rea'].value
        Rih = params['Rih'].value

        Ci = params['Ci'].value
        Ce = params['Ce'].value
        Ch = params['Ch'].value

        Ta = X['Ta']
        Ph = X['Ph']

        for i in range(1, num_rec):

            # the model equations
            dTi = ((Te[i-1] - Ti[i-1]) / (Rie * Ci) + (Th[i-1] - Ti[i-1]) / (Rih * Ci)) * self.rec_duration
            dTe = ((Ti[i-1] - Te[i-1]) / (Rie * Ce) + (Ta[i-1] - Te[i-1]) / (Rea * Ce)) * self.rec_duration
            dTh = ((Ti[i-1] - Th[i-1]) / (Rih * Ch) + (Ph[i-1]) / (Ch)) * self.rec_duration

            Ti[i] = Ti[i-1] + dTi
            Te[i] = Te[i-1] + dTe
            Th[i] = Th[i-1] + dTh

        return DarkGreyModelResult(Ti, X, params, {'Ti': Ti, 'Te': Te, 'Th': Th})


class TiTh(DarkGreyModel):
    '''
    A DarkGrey Model representing a TiTh RC-equivalent circuit

    Notes
    -----
    See "Bacher & Madsen (2011) Identifying suitable models for the heat dynamics of buildings.
    Energy and Buildings. 43. 1511-1522. 10.1016/j.enbuild.2011.02.005." for a complete description
    of RC thermal models and the eqiuvalent circuit diagram of TiTh.

    ~~~~
    # load data from e.g. pandas
    df = pd.read_csv()

    # assign internal temperature as the measured variable to be fitted to
    y = df['Internal Temperature [˚C]'].values

    # input values that will not change during the fit
    X = {
        'Ph': df['Boiler Power Output [kW]'].values,
        'Ta': df['Outside Air Temperature [˚C]'].values,
        'Th': df['Heating Circuit Temperature [˚C]'].values
    }

    # parameters to be fitted
    # 'value' - initial value
    # 'min' & 'max' - boundaries
    # 'vary' - if false, the parameter will be fixed to its initial value
    params = {
        'Ti0': {'value': y[0], 'vary': False, 'min': 15, 'max': 25},
        'Th0': {'value': y[0], 'vary': False, 'min': 10, 'max': 80},
        'Ci': {'value': 132},
        'Ch': {'value': 2.55, 'vary': False},
        'Ria': {'value': 1},
        'Rih': {'value': 0.65, 'vary': False}
    }

    # fit using the Nelder-Mead method
    model = TiTh(params, rec_duration=1).fit(X, y, method='nelder')
    ~~~~
    '''

    def model(self, params, X):
        '''
        The system of differential equations describing the model

        Parameters
        ----------
        params : `lmfit.Parameters`
            - 'Ti0' : Internal temperature at t(0)
            - 'Th0' : Heating system temperature at t(0)
            - 'Rih' : Thermal resistance between internal and heating system
            - 'Ria' : Thermal resistance between internal and ambient
            - 'Ci' : Thermal capacitance of internal
            - 'Ch' : Thermal capacitance of heating system
        X : dict
            - 'Ta' : List of ambient temperature values
            - 'Ph' : List of heating system power output values

        Returns
        -------
        Ti : np.array
            Fitted internal temperature values
        Th : np.array
            Fitted heating system temperature values
        '''

        num_rec = len(X['Ta'])

        Ti = np.zeros(num_rec)
        Th = np.zeros(num_rec)

        # alias these params/X so that the differential equations look pretty
        Ti[0] = params['Ti0']
        Th[0] = params['Th0']

        Rih = params['Rih'].value
        Ria = params['Ria'].value

        Ci = params['Ci'].value
        Ch = params['Ch'].value

        Ta = X['Ta']
        Ph = X['Ph']

        for i in range(1, num_rec):

            # the model equations
            dTi = ((Ta[i-1] - Ti[i-1]) / (Ria * Ci) + (Th[i-1] - Ti[i-1]) / (Rih * Ci)) * self.rec_duration
            dTh = ((Ti[i-1] - Th[i-1]) / (Rih * Ch) + (Ph[i-1]) / (Ch)) * self.rec_duration

            Ti[i] = Ti[i-1] + dTi
            Th[i] = Th[i-1] + dTh

        return DarkGreyModelResult(Ti, X, params, {'Ti': Ti, 'Th': Th})


class TiTe(DarkGreyModel):
    '''
    A DarkGrey Model representing a TiTe RC-equivalent circuit

    Notes
    -----
    See "Bacher & Madsen (2011) Identifying suitable models for the heat dynamics of buildings.
    Energy and Buildings. 43. 1511-1522. 10.1016/j.enbuild.2011.02.005." for a complete description
    of RC thermal models and the eqiuvalent circuit diagram of TiTeTh.

    ~~~~
    # load data from e.g. pandas
    df = pd.read_csv()

    # assign internal temperature as the measured variable to be fitted to
    y = df['Internal Temperature [˚C]'].values

    # input values that will not change during the fit
    X = {
        'Ph': df['Boiler Power Output [kW]'].values,
        'Ta': df['Outside Air Temperature [˚C]'].values,
    }

    # parameters to be fitted
    # 'value' - initial value
    # 'min' & 'max' - boundaries
    # 'vary' - if false, the parameter will be fixed to its initial value
    params = {
        'Ti0': {'value': y[0], 'vary': False, 'min': 15, 'max': 25},
        'Te0': {'value': y[0] - 2, 'vary': True, 'min': 10, 'max': 25},
        'Ci': {'value': 132},
        'Ce': {'value': 600},
        'Rie': {'value': 0.1},
        'Rea': {'value': 1},
    }

    # fit using the Nelder-Mead method
    model = TiTe(params, rec_duration=1).fit(X, y, method='nelder')
    ~~~~
    '''

    def model(self, params, X):
        '''
        The system of differential equations describing the model

        Parameters
        ----------
        params : `lmfit.Parameters`
            - 'Ti0' : Internal temperature at t(0)
            - 'Te0' : Thermal envelope temperature at t(0)
            - 'Rie' : Thermal resistance between internal and thermal envelope
            - 'Rea' : Thermal resistance between thermal envelope and ambient
            - 'Ci' : Thermal capacitance of internal
            - 'Ce' : Thermal capacitance of thermal envelope
        X : dict
            - 'Ta' : List of ambient temperature values
            - 'Ph' : List of heating system power output values

        Returns
        -------
        Ti : np.array
            Fitted internal temperature values
        Te : np.array
            Fitted thermal envelope temperature values
        '''

        num_rec = len(X['Ta'])

        Ti = np.zeros(num_rec)
        Te = np.zeros(num_rec)

        # alias these params/X so that the differential equations look pretty
        Ti[0] = params['Ti0']
        Te[0] = params['Te0']

        Rie = params['Rie'].value
        Rea = params['Rea'].value

        Ci = params['Ci'].value
        Ce = params['Ce'].value

        Ta = X['Ta']
        Ph = X['Ph']

        for i in range(1, num_rec):

            # the model equations
            dTi = ((Te[i-1] - Ti[i-1]) / (Rie * Ci) + (Ph[i-1]) / (Ci)) * self.rec_duration
            dTe = ((Ti[i-1] - Te[i-1]) / (Rie * Ce) + (Ta[i-1] - Te[i-1]) / (Rea * Ce)) * self.rec_duration

            Ti[i] = Ti[i-1] + dTi
            Te[i] = Te[i-1] + dTe

        return DarkGreyModelResult(Ti, X, params, {'Ti': Ti, 'Te': Te})


class Ti(DarkGreyModel):
    '''
    A DarkGrey Model representing a Ti RC-equivalent circuit

    Notes
    -----
    See "Bacher & Madsen (2011) Identifying suitable models for the heat dynamics of buildings.
    Energy and Buildings. 43. 1511-1522. 10.1016/j.enbuild.2011.02.005." for a complete description
    of RC thermal models and the eqiuvalent circuit diagram of Ti.

    ~~~~
    # load data from e.g. pandas
    df = pd.read_csv()

    # assign internal temperature as the measured variable to be fitted to
    y = df['Internal Temperature [˚C]'].values

    # input values that will not change during the fit
    X = {
        'Ph': df['Boiler Power Output [kW]'].values,
        'Ta': df['Outside Air Temperature [˚C]'].values,
    }

    # parameters to be fitted
    # 'value' - initial value
    # 'min' & 'max' - boundaries
    # 'vary' - if false, the parameter will be fixed to its initial value
    params = {
        'Ti0': {'value': y[0], 'vary': False, 'min': 15, 'max': 25},
        'Ci': {'value': 132},
        'Ria': {'value': 1},
    }

    # fit using the Nelder-Mead method
    model = Ti(params, rec_duration=1).fit(X, y, method='nelder')
    ~~~~
    '''

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.ic_param_names = ['Ti0']
        self.rc_param_names = ['Ci', 'Ria']
        self.input_param_names = ['Ta', 'Ph']

    def model(self, params, X):
        '''
        The system of differential equations describing the model

        Parameters
        ----------
        params : `lmfit.Parameters`
            - 'Ti0' : Internal temperature at t(0)
            - 'Ria' : Thermal resistance between internal and ambient
            - 'Ci' : Thermal capacitance of internal
        X : dict
            - 'Ta' : List of ambient temperature values
            - 'Ph' : List of heating system power output values

        Returns
        -------
        Ti : np.array
            Fitted internal temperature values
        '''

        num_rec = len(X['Ta'])

        Ti = np.zeros(num_rec)

        # alias these params/X so that the differential equations look pretty
        Ti[0] = params['Ti0'].value

        Ria = params['Ria'].value
        Ci = params['Ci'].value

        Ta = X['Ta']
        Ph = X['Ph']

        for i in range(1, num_rec):

            # the model equations
            dTi = ((Ta[i-1] - Ti[i-1]) / (Ria * Ci) + (Ph[i-1]) / (Ci)) * self.rec_duration

            Ti[i] = Ti[i-1] + dTi

        return DarkGreyModelResult(Ti, X, params, {'Ti': Ti})



from typing import Any, Callable, List, Optional, cast

import pandas as pd

from darkgreybox.train import train_models


def prefit_models(
    models: List[Any],
    X_train: pd.DataFrame,
    y_train: pd.Series,
    error_metric: Callable,
    prefit_splits: Optional[List] = None,
    prefit_filter: Optional[Callable] = None,
    method: str = 'nelder',
    obj_func: Optional[Callable] = None,
    n_jobs: int = -1,
    verbose: int = 10
) -> List[Any]:

    if prefit_splits is None:
        return models

    prefit_df = train_models(
        models=models,
        X_train=X_train,
        y_train=y_train,
        splits=prefit_splits,
        error_metric=error_metric,
        method=method,
        obj_func=obj_func,
        n_jobs=n_jobs,
        verbose=verbose
    )

    filtered_df = apply_prefit_filter(prefit_df, prefit_filter)

    if 'model' not in filtered_df or len(filtered_df.dropna(subset=['model'])) == 0:
        raise ValueError('No valid models found during prefit')

    return filtered_df['model'].tolist()


def apply_prefit_filter(prefit_df: pd.DataFrame, prefit_filter: Optional[Callable] = None) -> pd.DataFrame:
    """
    Applies the prefit filter to the prefit dataframe
    """
    if prefit_filter is None:
        return prefit_df
    else:
        return cast(pd.DataFrame, prefit_df[prefit_filter(prefit_df['error'])].reset_index(drop=True))



from timeit import default_timer as timer
from typing import Any, Callable, Dict, List, Union, cast

import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from darkgreybox import logger
from darkgreybox.base_model import DarkGreyModel, DarkGreyModelResult


def predict_models(
    models: List[Union[DarkGreyModel, Any]],
    X_test: pd.DataFrame,
    y_test: pd.Series,
    ic_params_map: Dict,
    error_metric: Callable,
    train_results: List[DarkGreyModelResult],
    n_jobs: int = -1,
    verbose: int = 10
) -> pd.DataFrame:
    """
    Generates the predictions for the `models` for the given `X_test` and `y_test` test data

    Params:
        models: list of `model.DarkGreyModel` objects
            list of models to be used for the predictions
        X_test: `pandas.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pandas.Series`
            A pandas Series of the test input data y
        ic_params_map: dict
            A dictionary of mapping functions that return the initial condition parameters
        error_metric: Callable
            An error metric function that confirms to the `sklearn.metrics` interface
        train_results: list of `model.DarkGreyModelResult`
            The model results of the previously trained models
        n_jobs: int
            The number of parallel jobs to be run as described by `joblib.Parallel`
        verbose: int
            The degree of verbosity as described by `joblib.Parallel`

    Returns:
        `pandas.DataFrame` with a record for each model's predictions

    Example:
    ~~~~

    from sklearn.metrics import mean_squared_error

    from darkgreybox.fit import test_models


    prefit_df = train_models(
        models=[trained_model_1, trained_model_2],
        X_test=X_test,
        y_test=y_test,
        ic_params_map={}
        error_metric=mean_squared_error,
        train_results=[trained_model_result_1, trained_model_result_2],
        n_jobs=-1,
        verbose=10
    )
    ~~~~
    """

    num_models = len(models)
    logger.info(f'Generating predictions for {num_models} models...')

    if n_jobs != 1:
        with Parallel(n_jobs=n_jobs, verbose=verbose) as p:
            df = cast(pd. DataFrame, pd.concat(
                cast(pd. DataFrame, p(delayed(predict_model)(
                    model, X_test, y_test, ic_params_map, error_metric, train_result)
                    for model, train_result in zip(models, train_results)
                )),
                ignore_index=True
            ))

    else:
        df = pd.concat([predict_model(model, X_test, y_test, ic_params_map, error_metric, train_result)
                        for model, train_result in zip(models, train_results)], ignore_index=True)

    return df


def predict_model(
    model: Union[DarkGreyModel, Any],
    X_test: pd.DataFrame,
    y_test: pd.Series,
    ic_params_map: Dict,
    error_metric: Callable,
    train_result: DarkGreyModelResult,
) -> pd.DataFrame:
    """
    Calculates redictions  of `model` for the given `X_test` and `y_test` test data.

    Params:
        model: `model.DarkGreyModel`
            model used for the prediction
        X_test: `pandas.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pandas.Series`
            A pandas Series of the test input data y
        ic_params_map: dict
            A dictionary of mapping functions that return the initial condition parameters
        error_metric: Callable
            An error metric function that confirms to the `sklearn.metrics` interface
        train_result: `model.DarkGreyModelResult`
            The model result of a previously trained model

    Returns:
        `pandas.DataFrame` with a single record for the fit model's predictions
    """

    start = timer()

    start_date = X_test.index[0]
    end_date = X_test.index[-1]

    X = X_test.to_dict(orient='list')
    y = y_test.values

    if isinstance(model, DarkGreyModel):

        ic_params = map_ic_params(ic_params_map, model, X_test, y_test, train_result)

        model_result = model.predict(X, ic_params)

        end = timer()

        return pd.DataFrame({
            'start_date': [start_date],
            'end_date': [end_date],
            'model': [model],
            'model_result': [model_result],
            'time': [end - start],
            'error': [error_metric(y, model_result.Z)]
        })

    else:
        end = timer()
        return pd.DataFrame({
            'start_date': [start_date],
            'end_date': [end_date],
            'model': [np.NaN],
            'model_result': [np.NaN],
            'time': [end - start],
            'error': [np.NaN]
        })


def map_ic_params(
    ic_params_map: Dict,
    model: DarkGreyModel,
    X_test: pd.DataFrame,
    y_test: pd.Series,
    train_result: DarkGreyModelResult
) -> Dict:
    """
    Maps the test initial condition parameters according to `ic_params_map`

    Parameters:
        ic_params_map: Dict
            A dictionary of mapping functions that return the initial condition parameters
        model: `model.DarkGreyModel`
            model used for the prediction
        X_test: `pd.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pd.Series`
            A pandas Series of the test input data y
        train_results: `model.DarkGreyModelResult`
            model result object for training data

    Returns the initial conditions parameters dict

    ~~~~
    Assuming y_test holds the internal temperatures `Ti`

    ic_params_map = {
        'Ti0': lambda X_test, y_test, train_result: y_test.iloc[0],
        'Th0': lambda X_test, y_test, train_result: y_test.iloc[0],
        'Te0': lambda X_test, y_test, train_result: train_result.Te[-1],
    }

    will map the first internal temperature in the test set to both `Ti0` and `Th0`
    and the last `Te` value from the training results to `Te0`
    ~~~~
    """

    ic_params = {}

    for key in ic_params_map:
        if key in model.params:
            ic_params[key] = ic_params_map[key](X_test, y_test, train_result)

    return ic_params
